module rad_rte_two_stream_app_2d_wrapper

  use vtype_module
  use rad_rte_two_stream_app_v3, only : &
    & IDScatApproxEddington, &
    & IDScatApproxHemiMean

  implicit none

  private

  public :: RadRTETwoStreamApp2DWrapper
  public :: RadRTETwoStreamApp2DLWWrapper
  public :: RadRTETwoStreamApp2DSWWrapper
  public :: RadRTETwoStreamApp2DWrapperInit

  public :: IDScatApproxEddington
  public :: IDScatApproxHemiMean


  integer, parameter :: jmax = 1



  character(128), save :: ModuleName = 'rad_rte_two_stream_app_2d_wrapper'

  logical, save :: FlagInited = .false.

contains

  !----------------------------------------------------------------------------

  subroutine RadRTETwoStreamApp2DWrapper(           &
    & IDScatApprox, FlagSrcFuncTech,                & ! (in)
    & imax, kmax,                                   & ! (in)
    & xz_SSA, xz_AF,                                & ! (in)
    & xr_OptDep,                                    & ! (in)
    & x_SurfAlbedo,                                 & ! (in)
    & xr_RadUwFlux, xr_RadDwFlux,                   & ! (out)
    & SolarFluxTOA, x_InAngle,                      & ! (in)
    & xr_PFInted, x_SurfPFInted, x_SurfDPFDTInted,  & ! (in)
    & xra_DelRadUwFlux, xra_DelRadDwFlux            & ! (out)
    & )

    use rad_rte_two_stream_app_v3, only : RadRTETwoStreamAppWrapper

    integer , intent(in ) :: IDScatApprox
    logical , intent(in ) :: FlagSrcFuncTech
    integer , intent(in ) :: imax
    integer , intent(in ) :: kmax
    real(DP), intent(in ) :: xz_SSA          (1:imax,1:kmax)
    real(DP), intent(in ) :: xz_AF           (1:imax,1:kmax)
    real(DP), intent(in ) :: xr_OptDep       (1:imax,0:kmax)
    real(DP), intent(in ) :: x_SurfAlbedo    (1:imax)
    real(DP), intent(out) :: xr_RadUwFlux    (1:imax,0:kmax)
    real(DP), intent(out) :: xr_RadDwFlux    (1:imax,0:kmax)

    real(DP), intent(in ) :: SolarFluxTOA
    real(DP), intent(in ) :: x_InAngle       (1:imax)
    real(DP), intent(in ) :: xr_PFInted      (1:imax,0:kmax)
    real(DP), intent(in ) :: x_SurfPFInted   (1:imax)
    real(DP), intent(in ) :: x_SurfDPFDTInted(1:imax)
    real(DP), intent(out) :: xra_DelRadUwFlux(1:imax,0:kmax,-1:1)
    real(DP), intent(out) :: xra_DelRadDwFlux(1:imax,0:kmax,-1:1)


    ! Local variables
    !
    logical, parameter :: FlagTOAFlux = .true.
    logical, parameter :: FlagEmis    = .true.

    real(DP) :: xyz_SSA       ( 0:imax-1, 1:jmax, 1:kmax )
    real(DP) :: xyz_AF        ( 0:imax-1, 1:jmax, 1:kmax )
    real(DP) :: xyr_OptDep    ( 0:imax-1, 1:jmax, 0:kmax)
    real(DP) :: xy_SurfAlbedo ( 0:imax-1, 1:jmax )
    real(DP) :: xyr_RadUwFlux(0:imax-1, 1:jmax, 0:kmax)
    real(DP) :: xyr_RadDwFlux(0:imax-1, 1:jmax, 0:kmax)

    real(DP) :: xy_InAngle       (0:imax-1, 1:jmax)
    real(DP) :: xyr_PFInted      (0:imax-1, 1:jmax, 0:kmax)
    real(DP) :: xy_SurfPFInted   (0:imax-1, 1:jmax)
    real(DP) :: xy_SurfDPFDTInted(0:imax-1, 1:jmax)
    real(DP) :: xyra_DelRadUwFlux(0:imax-1, 1:jmax, 0:kmax, -1:1)
    real(DP) :: xyra_DelRadDwFlux(0:imax-1, 1:jmax, 0:kmax, -1:1)

    integer :: k
    integer :: l


    if ( .not. FlagInited ) then
      write( 6, * ) trim( ModuleName ), " is not initialized."
      stop
    end if


    do k = 1, kmax
      xyz_SSA(:,1,k) = xz_SSA(:,k)
      xyz_AF (:,1,k) = xz_AF (:,k)
    end do
    do k = 0, kmax
      xyr_OptDep(:,1,k) = xr_OptDep(:,k)
    end do
    xy_SurfAlbedo(:,1) = x_SurfAlbedo

    xy_InAngle(:,1) = x_InAngle
    do k = 0, kmax
      xyr_PFInted(:,1,k) = xr_PFInted(:,k)
    end do
    xy_SurfPFInted   (:,1) = x_SurfPFInted
    xy_SurfDPFDTInted(:,1) = x_SurfDPFDTInted



    call RadRTETwoStreamAppWrapper(                &
      & IDScatApprox, FlagTOAFlux, FlagEmis, FlagSrcFuncTech,  & ! (in)
      & imax, jmax, kmax,                                & ! (in)
      & xyz_SSA, xyz_AF,                                 & ! (in)
      & xyr_OptDep,                                      & ! (in)
      & xy_SurfAlbedo,                                   & ! (in)
      & xyr_RadUwFlux, xyr_RadDwFlux,                    & ! (out)
      & SolarFluxTOA, xy_InAngle,                        & ! (in) optional
      & xyr_PFInted, xy_SurfPFInted, xy_SurfDPFDTInted,  & ! (in) optional
      & xyra_DelRadUwFlux, xyra_DelRadDwFlux             & ! (out) optional
      & )


    do k = 0, kmax
      xr_RadUwFlux(:,k) = xyr_RadUwFlux(:,1,k)
      xr_RadDwFlux(:,k) = xyr_RadDwFlux(:,1,k)
    end do
    do l = -1, 1
      do k = 0, kmax
        xra_DelRadUwFlux(:,k,l) = xyra_DelRadUwFlux(:,1,k,l)
        xra_DelRadDwFlux(:,k,l) = xyra_DelRadDwFlux(:,1,k,l)
      end do
    end do


  end subroutine RadRTETwoStreamApp2DWrapper

  !----------------------------------------------------------------------------

  subroutine RadRTETwoStreamApp2DLWWrapper(         &
    & imax, kmax,                                   & ! (in)
    & xz_SSA, xz_AF,                                & ! (in)
    & xr_OptDep,                                    & ! (in)
    & x_SurfAlbedo,                                 & ! (in)
    & xr_PFInted, xr_DPFDTInted,                    & ! (in)
    & x_SurfPFInted, x_SurfDPFDTInted,              & ! (in)
    & xr_RadUwFlux, xr_RadDwFlux,                   & ! (out)
    & xra_DelRadUwFlux, xra_DelRadDwFlux            & ! (out)
    & )

    ! USE statements
    !
    use rad_rte_two_stream_app_v3, only : RadRTETwoStreamAppLW


    integer , intent(in ) :: imax
    integer , intent(in ) :: kmax
    real(DP), intent(in ) :: xz_SSA          (1:imax,1:kmax)
    real(DP), intent(in ) :: xz_AF           (1:imax,1:kmax)
    real(DP), intent(in ) :: xr_OptDep       (1:imax,0:kmax)
    real(DP), intent(in ) :: x_SurfAlbedo    (1:imax)
    real(DP), intent(in ) :: xr_PFInted      (1:imax,0:kmax)
    real(DP), intent(in ) :: xr_DPFDTInted   (1:imax,0:kmax)
    real(DP), intent(in ) :: x_SurfPFInted   (1:imax)
    real(DP), intent(in ) :: x_SurfDPFDTInted(1:imax)
    real(DP), intent(out) :: xr_RadUwFlux    (1:imax,0:kmax)
    real(DP), intent(out) :: xr_RadDwFlux    (1:imax,0:kmax)
    real(DP), intent(out) :: xra_DelRadUwFlux(1:imax,0:kmax, -1:1)
    real(DP), intent(out) :: xra_DelRadDwFlux(1:imax,0:kmax, -1:1)


    ! Local variables
    !
    real(DP) :: xyz_SSA          (0:imax-1, 1:jmax, 1:kmax)
    real(DP) :: xyz_AF           (0:imax-1, 1:jmax, 1:kmax)
    real(DP) :: xyr_OptDep       (0:imax-1, 1:jmax, 0:kmax)
    real(DP) :: xy_SurfAlbedo    (0:imax-1, 1:jmax)
    real(DP) :: xyr_PFInted      (0:imax-1, 1:jmax, 0:kmax)
    real(DP) :: xyr_DPFDTInted   (0:imax-1, 1:jmax, 0:kmax)
    real(DP) :: xy_SurfPFInted   (0:imax-1, 1:jmax)
    real(DP) :: xy_SurfDPFDTInted(0:imax-1, 1:jmax)
    real(DP) :: xyr_RadUwFlux    (0:imax-1, 1:jmax, 0:kmax)
    real(DP) :: xyr_RadDwFlux    (0:imax-1, 1:jmax, 0:kmax)
    real(DP) :: xyra_DelRadUwFlux(0:imax-1, 1:jmax, 0:kmax, -1:1)
    real(DP) :: xyra_DelRadDwFlux(0:imax-1, 1:jmax, 0:kmax, -1:1)

    integer :: k
    integer :: l


    if ( .not. FlagInited ) then
      write( 6, * ) trim( ModuleName ), " is not initialized."
      stop
    end if


    do k = 1, kmax
      xyz_SSA    (:,1,k) = xz_SSA    (:,k)
      xyz_AF     (:,1,k) = xz_AF     (:,k)
    end do
    do k = 0, kmax
      xyr_OptDep    (:,1,k) = xr_OptDep    (:,k)
      xyr_PFInted   (:,1,k) = xr_PFInted   (:,k)
      xyr_DPFDTInted(:,1,k) = xr_DPFDTInted(:,k)
    end do
    xy_SurfAlbedo    (:,1) = x_SurfAlbedo
    xy_SurfPFInted   (:,1) = x_SurfPFInted
    xy_SurfDPFDTInted(:,1) = x_SurfDPFDTInted

    call RadRTETwoStreamAppLW(                     &
      & imax, jmax, kmax,                                & ! (in)
      & xyz_SSA, xyz_AF,                                 & ! (in)
      & xyr_OptDep,                                      & ! (in)
      & xy_SurfAlbedo,                                   & ! (in)
      & xyr_PFInted, xyr_DPFDTInted,                     & ! (in)
      & xy_SurfPFInted, xy_SurfDPFDTInted,               & ! (in)
      & xyr_RadUwFlux, xyr_RadDwFlux,                    & ! (out)
      & xyra_DelRadUwFlux, xyra_DelRadDwFlux             & ! (out)
      & )

    do k = 0, kmax
      xr_RadUwFlux(:,k) = xyr_RadUwFlux(:,1,k)
      xr_RadDwFlux(:,k) = xyr_RadDwFlux(:,1,k)
    end do
    do l = -1, 1
      do k = 0, kmax
        xra_DelRadUwFlux(:,k,l) = xyra_DelRadUwFlux(:,1,k,l)
        xra_DelRadDwFlux(:,k,l) = xyra_DelRadDwFlux(:,1,k,l)
      end do
    end do


  end subroutine RadRTETwoStreamApp2DLWWrapper

  !------------------------------------------------------------------------------------

  subroutine RadRTETwoStreamApp2DSWWrapper(      &
    & imax, kmax,                     & ! (in)
    & xz_SSA, xz_AF,                  & ! (in)
    & xr_OptDep,                      & ! (in)
    & x_SurfAlbedo,                   & ! (in)
    & SolarFluxTOA, x_InAngle,        & ! (in)
    & xr_RadUwFlux, xr_RadDwFlux      & ! (out)
    & )

    ! USE statements
    !
    use rad_rte_two_stream_app_v3, only : RadRTETwoStreamAppSW

    integer , intent(in ) :: imax
    integer , intent(in ) :: kmax
    real(DP), intent(in ) :: xz_SSA        (1:imax,1:kmax)
    real(DP), intent(in ) :: xz_AF         (1:imax,1:kmax)
    real(DP), intent(in ) :: xr_OptDep     (1:imax,0:kmax)
    real(DP), intent(in ) :: x_SurfAlbedo  (1:imax )
    real(DP), intent(in ) :: SolarFluxTOA
    real(DP), intent(in ) :: x_InAngle     (1:imax)
    real(DP), intent(out) :: xr_RadUwFlux  (1:imax,0:kmax)
    real(DP), intent(out) :: xr_RadDwFlux  (1:imax,0:kmax)


    ! Local variables
    !
    real(DP) :: xyz_SSA      ( 0:imax-1, 1:jmax, 1:kmax )
    real(DP) :: xyz_AF       ( 0:imax-1, 1:jmax, 1:kmax )
    real(DP) :: xyr_OptDep   ( 0:imax-1, 1:jmax, 0:kmax )
    real(DP) :: xy_SurfAlbedo( 0:imax-1, 1:jmax )
    real(DP) :: xy_InAngle   ( 0:imax-1, 1:jmax )
    real(DP) :: xyr_RadUwFlux( 0:imax-1, 1:jmax, 0:kmax )
    real(DP) :: xyr_RadDwFlux( 0:imax-1, 1:jmax, 0:kmax )

    integer :: k


    ! 初期化確認
    ! Initialization check
    !
    if ( .not. FlagInited ) then
      write( 6, * ) "Error: ", trim( ModuleName ), &
        & 'This module has not been initialized.'
      stop
    end if


    do k = 1, kmax
      xyz_SSA(:,1,k)      = xz_SSA(:,k)
      xyz_AF (:,1,k)      = xz_AF (:,k)
    end do
    do k = 0, kmax
      xyr_OptDep(:,1,k)   = xr_OptDep(:,k)
    end do
    xy_SurfAlbedo(:,1)    = x_SurfAlbedo
    xy_InAngle   (:,1)    = x_InAngle

    call RadRTETwoStreamAppSW(      &
      & imax, jmax, kmax,                 & ! (in)
      & xyz_SSA, xyz_AF,                  & ! (in)
      & xyr_OptDep,                       & ! (in)
      & xy_SurfAlbedo,                    & ! (in)
      & SolarFluxTOA, xy_InAngle,         & ! (in)
      & xyr_RadUwFlux, xyr_RadDwFlux      & ! (out)
      & )

    do k = 0, kmax
      xr_RadUwFlux(:,k) = xyr_RadUwFlux(:,1,k)
      xr_RadDwFlux(:,k) = xyr_RadDwFlux(:,1,k)
    end do


  end subroutine RadRTETwoStreamApp2DSWWrapper

  !----------------------------------------------------------------------------

  subroutine RadRTETwoStreamApp2DWrapperInit

    use rad_rte_two_stream_app_v3, only : RadRTETwoStreamAppInit


    if ( FlagInited ) return


    call RadRTETwoStreamAppInit


    FlagInited = .true.

  end subroutine RadRTETwoStreamApp2DWrapperInit

  !----------------------------------------------------------------------------

end module rad_rte_two_stream_app_2d_wrapper

